#include <rtlib.hpp>
#include <rtlib/taskstat.hpp>
#include <stdio.h>
#include <rtlib/cpu.hpp>
#include <stdlib.h>
#include <math.h>

#include <fstream>
#include <iomanip>
#include <string>
#include <rtlib/converter.hpp>
#include <exception>

using namespace RTSim;
using namespace Simulation;
using namespace MetaSim;

#define DEB(A) A
#define DDB(A)
#define UTIL_PRECISION 0.02
#define MIN_DOUBLE_CONST 2e-13
#define STATE_FNAME ".medf_state"

// the lateness of a job is defined as max{0, f - d}, where f and d
// are, respectively, the finish time and the absolute deadline of the job
typedef LatenessStat<StatMax> maxLat;
typedef LatenessStat<StatMean> meanLat;


#if 0
// log files
struct Out_t {
      ofstream max; // max tardiness
      ofstream mean; // mean tardiness
} ;
#endif

int seed; // seed for random number generation

/*
 * If the following parameter is true, then heavy tasks have
 * a small execution time and hence a short period, whereas
 * light tasks have a large execution time and hence a very
 * large period.
 */
const bool worstCaseScenario = false ; // experimental ...

/* 
 * The parameters of the simulations are: minimum number minCPU of
 * CPUs, maximum number maxCPU of CPUs, maximum number nIter of
 * iterations per CPU number, and duration dur of each
 * iteration. Simulations are carried out as follows.  A number of
 * systems, equal to (maxCPU - minCPU + 1), and with, respectively,
 * minCPU, minCPU+1, ..., and maxCPU CPUs are simulated. For each
 * system, nIter different sporadic task sets with relative deadlines
 * coinciding with the periods, and total utilization equal to the
 * number of CPUs multiplied by a factor totUcapRatio are generated.
 * We define target total utilization the just defined product.  To
 * get maximum precision, totUcapRatio must be an integer multiple of
 * the inverse of the quantity scale (described in detail in next
 * comments).
 */
int minCPU = 2 ; // minimum number of CPUs
int maxCPU = 2 ; // maximum number of CPUs
int nIter = 1 ;  // maximum number of iterations per CPU number
// ratio between the target total utilization and the total system capacity 
double totUcapRatio = 1 ; 
int M = 1 ; // number of CPUs in the system currently being simulated
int iter = 1 ; // current simulation run

/*
 * We define a task as heavy or light, if it has a utilization, respectively,
 * higher than or strictly lower than a fixed value Uminheavy.
 */
double Uminheavy = 0.4 ;
//double Umaxheavy = 0.999 ;
double Umaxheavy = 0.99 ;
double Umaxlight = Uminheavy ; // will be changed in worst-case scenario

/*
 * The following technique is adopted to generate each task set so as to 
 * avoid approximation errors.
 *
 * An integer constant larger than 1 and called scale is defined.  The
 * utilization U of each task is represented as a natural number U_rep
 * ranging from 0 to scale, and assuming that U = U_rep / scale.  To
 * get a reasonable precision for the representation of the
 * utilization, we set scale to 1000.
 *
 * Task periods are random integer multiples of scale.  As a
 * consequence, the execution time, equal to (period * U), of the jobs
 * of any task is guaranteed to be an integer. In a previous version
 * of the code: the minimum possible execution time of each job was
 * guaranteed to be in the order of the minimum possible period,
 * i.e. in the order of scale. To this purpose each task period was
 * forced to not be smaller than scale/utilization. This gave rise to
 * a worst-case minimum period equal to scale/Umin, where Umin is the
 * minimum task utilization that can be generated. I do not remember
 * the reason for this guarantee, and I removed it.
 */
const int scale = 1000 ; // scaling factor to reduce approximation errors
// representation of the minimum possible randomly generated utilization
const int Umin_rep = scale/100 ;  

/*
 * We continue now the description of the generation of the task
 * parameters. First, a set (possibly empty, see below) of heavy tasks
 * is generated. The utilization of each such task is a random number,
 * uniformly distributed between Uminheavy and Umaxheavy. Hence, each
 * heavy task has a mean utilization (1 + Uminheavy)/2 (see the code
 * for details on rounding off errors).  We impose that the sum of the
 * mean utilizations of heavy tasks not be higher than a fraction
 * heavyFraction <= 1 of the target total capacity.  Hence assigning 0
 * to heavyFraction means generating no heavy task.
 *
 * Then, light tasks are generated so as to fill the remaining
 * capacity.  To this purpose, each light task must have a mean
 * utilization Umeanlight = (targetUtot-totalUheavy)/nlight, where
 * totalUheavy is the sum of the utilizations of the nheavy previously
 * generated heavy tasks, targetUtot is equal to
 * num_processors*totUcapRatio and nlight is the number of light
 * tasks.  The utilization of the light tasks is a random number
 * between Umin and a maximum value Umaxlight.  Thus the mean
 * utilization of each light tasks is equal to (Umaxlight+Umin)/2. To
 * fill the remaining system capacity, such quantity must be equal to
 * the previously computed value Umeanlight.  As a consequence, we get
 * Umaxlight = 2*Umeanlight - Umin.  But Umaxlight can not be higher
 * than Uminheavy. In order for such relation to hold, it is easy to
 * prove that a sufficient condition is nlight >= minlight =
 * 2*(targetUtot-totalHeavy)/(Umin+Uheavy-1/scale).  nlight is
 * actually nitially set to lightRatio*minlight, with lightRatio >= 1.
 */

/* Note: for the task generation to work correctly, the representation
 * of the minimum possible utilization for a heavy task must must be
 * at least twice as large as Umin_rep.
 */

/*
 * Maximum fraction of the total capacity allocated to heavy tasks. If
 * worst-case mode is on, the heavy fraction will be automatically
 * increased so as to guarantee that it is at least (M-1)/M
 */
double heavyFraction = .99 ;

//double heavyFraction = 1 ;
double lightRatio = 1.1 ; // it must be no lower than 1

/*
 * Maximum possible normalized period, with respect to scale, for a
 * task. In other words, the maximum possible period for a task is
 * equal to maxNormPeriod * scale. We define the parameter
 * maxNormPeriod, because, using this parameter, it is easier to
 * generate periods that are integer multiples of scale.
 */
int maxNormPeriod = 100 ;

/*
 * Similarly, we define the minimum normalized period
*/
int minNormPeriod = 1 ;

/*
 * For each system and each task set, the schedule produced by EDF is
 * simulated for dur ticks. To let all tasks be scheduled at least
 * once, the duration of each simulation run must be equal to at least the
 * maximum possible task period, i.e., maxNormPeriod * scale.
 * To make it easy to set the duration to a multiple of this quantity,
 * we define the following global variable, dur, equal to the duration
 * of each run, measured in multiples of maxNormPeriod * scale.
 */
Tick dur = 21000 ;

/*
 * Still about duration, rtsim supports at most the following maximum duration.
 * Such a value is used to cap the actual duration of each run so that
 * rtsim works correctly.
 */
Tick maxSupportedDur = (21000ULL * 100 * 1000) ;

/*
 * The maximum per-task fluctuation of inter-arrival times is
 * controlled trhough the following sporad_coeff parameter. In more
 * detail, inter-arrival times are uniformly distributed in
 * [period, period * (1 + sporad_coeff)]
 */
const double sporad_coeff = 0.01 ;

// All statistic are stored in variables of type per_task_stat, which is a 
// tuple of information for a given task
struct per_task_stat {
  int task ;        // index of the task in the task set
  int iter ;        // iteration in which the task has been generated
  int CPU ;         // number of CPUs in the system
  int exec_time ;   // task execution time
  int max_exec_time;// maximum execution time in the task set
  double th_bound ; // theoretical lateness upper bound
  double th_bound_UJ ; // theoretical lateness upper bound from Uma and James
  double maxlat ;   // maximum lateness experienced by the task
  double meanlat ;  // mean lateness experienced by the task
  double diff ;     // th_bound - max_lat
  double ratio ;    // max_lat/th_bound
  double ratio_UJ;    // max_lat/th_bound_UJ
} ;

/*
     * During the simulation several aggregated statistics will be computed and
     * stored inside variables of type per_task_stat.
     * Their names have the form: measure_aggregation_level.
     *
     * The possible aggregation levels are:
     * "": aggregated value computed over all the tasks for a single iteration
     * "per_CPU": aggregated value computed over all the iterations for a system
     * "global": aggregated value computed over all the simulated systems
     *
     * The measures recorded are:
     *
     * [min, max]diff: minimum/maximum difference between the theoretical 
     *                 upper bound for the lateness of a task and its actual 
     *                 maximum lateness
     * [min, max]ratio: minimum/maximum ratio between the theoretical 
     *                  upper bound for the lateness of a task and its actual 
     *                  maximum lateness
     * maxratio_UJ: maximum ratio between the theoretical 
     *                  upper bound as computed by Uma and James (UJ) for the 
     *                  lateness of a task and its actual maximum lateness
     *
     */

    per_task_stat mindiff_global = { 
                           0,                    // task
                           0,                    // iter
                           0,                    // number of CPUs in the system
                           0,                    // task execution time
                           0,                    // maximum execution time
                           0,                    // th. upper bound
                           0,                    // th. upper bound UJ
                           0,                    // max lateness experienced
                           0,                    // mean lateness experienced
                           -MIN_DOUBLE_CONST -1, // diff
                           0,                    // ratio
			   0                     // ratio_UJ
                          } ;
    per_task_stat maxdiff_global = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;
    per_task_stat minratio_global = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0 } ;
    per_task_stat maxratio_global = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;
    per_task_stat maxratio_UJ_global = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;
    per_task_stat maxlat_global = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;

    // Per system statistics are stored in four per_task_stat arrays, which 
    // will store the data corresponding to the task that experienced, 
    // respectively, mindiff, maxdiff, minratio and maxratio over all the 
    // iterations.

    per_task_stat *mindiff_per_CPU, *maxdiff_per_CPU,
      *minratio_per_CPU, *maxratio_per_CPU, *maxratio_UJ_per_CPU,
      *maxlat_per_CPU;

// Four helper functions for saving the state at each iteration and loading 
// the last state saved in case of a restart
void save_task_stat(ofstream &f, const per_task_stat *s)
{
  f<<s->task<<endl ;         // index of the task in the task set
  f<<s->iter<<endl ;         // iteration in which the task has been generated
  f<<s->CPU<<endl ;          // number of CPUs in the system
  f<<s->exec_time<<endl ;    // task execution time
  f<<s->max_exec_time<<endl ;// maximum execution time in the task set
  f<<s->th_bound<<endl ;     // theoretical lateness upper bound
  f<<s->th_bound_UJ<<endl ;  // theoretical lateness upper bound UJ
  f<<s->maxlat<<endl ;       // maximum lateness experienced by the task
  f<<s->meanlat<<endl ;      // mean lateness experienced by the task
  f<<s->diff<<endl ;         // th_bound - max_lat
  f<<s->ratio<<endl ;        // max_lat/th_bound
  f<<s->ratio_UJ<<endl ;     // max_lat/th_bound_UJ
}

void load_task_stat(ifstream &f, per_task_stat *s)
{
  f>>s->task ;         // index of the task in the task set
  f>>s->iter ;         // iteration in which the task has been generated
  f>>s->CPU ;          // number of CPUs in the system
  f>>s->exec_time ;    // task execution time
  f>>s->max_exec_time ;// maximum execution time in the task set
  f>>s->th_bound ;     // theoretical lateness upper bound
  f>>s->th_bound_UJ ;  // theoretical lateness upper bound from Uma and James
  f>>s->maxlat ;       // maximum lateness experienced by the task
  f>>s->meanlat ;      // mean lateness experienced by the task
  f>>s->diff ;         // th_bound - max_lat
  f>>s->ratio ;        // max_lat/th_bound
  f>>s->ratio_UJ ;     // max_lat/th_bound_UJ
}

void save_state()
{
  DDB(cout<<"save_state: opening file ..."<<endl) ;
  ofstream f(STATE_FNAME) ;
  if (f.fail()) {
    cerr<<"save_state: errore nell'apertura del file "<<STATE_FNAME<<endl ;
    exit(-1) ;
  }
  f<<rand()<<endl ;
  f<<minCPU<<endl ;
  f<<maxCPU<<endl ;
  f<<nIter<<endl ;
  f<<totUcapRatio<<endl ;
  f<<Uminheavy<<endl ;
  f<<heavyFraction<<endl ;
  f<<lightRatio<<endl ;
  f<<dur/(maxNormPeriod*scale)<<endl ;
  f<<M<<endl ;
  f<<iter<<endl<<endl ; // just terminated iteration
  DDB(cout<<"save_state: written first set of data ..."<<endl) ;
  per_task_stat * const glob[6] = {&mindiff_global, &maxdiff_global, 
				   &minratio_global, &maxratio_global, 
				   &maxratio_UJ_global,
				   &maxlat_global} ;
  per_task_stat * const per_CPU[6] = {mindiff_per_CPU, maxdiff_per_CPU,
				      minratio_per_CPU, maxratio_per_CPU, 
				      maxratio_UJ_per_CPU,
				      maxlat_per_CPU} ;
  for (int i = 0 ; i < 6 ; i++) {
    save_task_stat(f, glob[i]) ;
    DDB(cout<<"save_state: just saved glob["<<i<<"]"<<endl) ;
    for (int j = minCPU-1 ; j < M ; j++) {
      DDB(cout<<"save_state: saving per_CPU["<<i<<"]["<<j<<"]"<<endl) ;
      DDB(cout<<"per_CPU["<<i<<"]["<<j<<"] = "<<&per_CPU[i][j]<<endl) ;
      save_task_stat(f, &per_CPU[i][j]) ;
      DDB(cout<<"save_state: just saved per_CPU["<<i<<"]["<<j<<"]"<<endl) ;
    }
  }
  DDB(cout<<"save_state: written all data ..."<<endl) ;
  f.close() ;
}

void load_state() 
{
  DEB(cout<<"Loading state ..."<<endl) ;
  ifstream f(STATE_FNAME) ;
  if (f.fail()) {
    cerr<<"Error in opening state file "<<STATE_FNAME<<endl ;
    exit(-1) ;
  }
  f>>seed ;
  f>>minCPU ;
  f>>maxCPU ;
  f>>nIter ;
  f>>totUcapRatio ;
  f>>Uminheavy ;
  f>>heavyFraction ;
  f>>lightRatio ;
  unsigned int dur_i ;
  f>>dur_i ; // >> operator lacking for type Tick
  dur = dur_i ;
  f>>M ; // system on which last run stopped
  f>>iter ;  // last iteration performed before last run stoppped
  DEB(cout<<"load_state: loaded first set of data ..."<<endl) ;

  mindiff_per_CPU = new per_task_stat[maxCPU] ;
  maxdiff_per_CPU = new per_task_stat[maxCPU] ;
  minratio_per_CPU = new per_task_stat[maxCPU] ;
  maxratio_per_CPU = new per_task_stat[maxCPU] ;
  maxratio_UJ_per_CPU = new per_task_stat[maxCPU] ;
  maxlat_per_CPU = new per_task_stat[maxCPU] ;

  per_task_stat * const glob[6] = {&mindiff_global, &maxdiff_global, 
				   &minratio_global, &maxratio_global, 
				   &maxratio_UJ_global,
				   &maxlat_global} ;
  per_task_stat * const per_CPU[6] = {mindiff_per_CPU, maxdiff_per_CPU,
				      minratio_per_CPU, maxratio_per_CPU,
				      maxratio_UJ_per_CPU,
				      maxlat_per_CPU} ;
  for (int i = 0 ; i < 6 ; i++) {
    load_task_stat(f, glob[i]) ;
    for (int j = minCPU-1 ; j < M ; j++)
      load_task_stat(f, &per_CPU[i][j]) ;
  }
  f.close() ;
  iter ++ ; // move to next iteration (may cause moving to next system)
  DEB(cout<<"load_state finished"<<endl) ;
}

void interactive_exit() {
    cerr<<"\n[P]roceed or [E]xit? " ;
    char c ;
    cin>>c ;
    if (c == 'P' || c == 'p')
	return ;
    exit(-1) ;
}

// print per-system summary
void per_system_summaries(
#if 0
Out_t &Out
#endif
) {
#if 0
       cout<<"\nPer system summary on mindiff:\n"
                 "[nCPU]\t\t[th. bound]\t\t[max_lateness]\t\t[mean_lateness]"<<
		 endl ;
//#if 0
       Out.max<<"\nPer system summary on mindiff:\n"
                "[nCPU]\t[th. bound]\t[max_lat]\t[mean_lat]"<<endl ;
//#endif
       for(int i = minCPU - 1 ; i < M ; i++) {
              cout<<i+1<<"\t\t"
		      <<setw(11)<<mindiff_per_CPU[i].th_bound<<"\t\t"
                      <<setw(11)<<mindiff_per_CPU[i].maxlat<<"\t\t"
                      <<setw(11)<<mindiff_per_CPU[i].meanlat<<endl ;
//#if 0         
	      Out.max<<i+1<<"\t"<<mindiff_per_CPU[i].th_bound<<"\t"
                      <<mindiff_per_CPU[i].maxlat<<"\t"
                      <<mindiff_per_CPU[i].meanlat<<endl ;
//#endif
       }

       cout<<"\nPer system summary on maxratio:\n"
                 "[nCPU]\t[th_bound]\t\t[max_ratio]\t\t[mean_ratio]"<<endl ;
//#if 0
       Out.max<<"\nPer system summary on maxratio:\n"
                 "[nCPU]\t[max_ratio]\t[mean_ratio]"<<endl ;
//#endif
       for(int i = minCPU - 1 ; i < M ; i++) {
           cout<<i+1<<"\t"
	       <<setw(11)<<maxratio_per_CPU[i].th_bound<<"\t\t"
	       <<setw(11)<<maxratio_per_CPU[i].ratio<<"\t\t"
	       <<setw(11)
	       <<maxratio_per_CPU[i].meanlat/maxratio_per_CPU[i].th_bound
	       <<endl ;
//#if 0
           Out.max<<i+1<<"\t"<<maxratio_per_CPU[i].ratio<<"\t"
                  <<maxratio_per_CPU[i].meanlat/maxratio_per_CPU[i].th_bound
                  <<endl ;
//#endif
       }
       cout<<"\nPer system summary on maxratio_UJ:\n"
	 "[nCPU]\t[th_bound_UJ]\t\t[max_ratio_UJ]\t\t[mean_ratio_UJ]"<<endl ;
       for(int i = minCPU - 1 ; i < M ; i++) {
           cout<<i+1<<"\t"
	       <<setw(11)<<maxratio_UJ_per_CPU[i].th_bound_UJ<<"\t\t"
	       <<setw(11)<<maxratio_UJ_per_CPU[i].ratio_UJ<<"\t\t"
	       <<setw(11)
	       <<maxratio_UJ_per_CPU[i].meanlat/
	         maxratio_UJ_per_CPU[i].th_bound_UJ
	       <<endl ; 
       }
#endif

       cout<<"\nPer system summary on max lateness:\n"
                 "[nCPU]\t\t[max_lateness]\t\t[mean_lateness]"<<endl ;
#if 0
       Out.max<<"\nPer system summary on max lateness:\n"
                 "[nCPU]\t[max_lateness]\t[mean_lateness]"<<endl ;
#endif
       for(int i = minCPU - 1 ; i < M ; i++) {
           cout<<i+1<<"\t\t"<<setw(11)<<maxlat_per_CPU[i].maxlat<<"\t\t"
                   <<setw(11)<<maxlat_per_CPU[i].meanlat<<endl ;
#if 0
           Out.max<<i+1<<"\t"<<maxlat_per_CPU[i].maxlat<<"\t"
                  <<maxlat_per_CPU[i].meanlat<<endl ;
#endif
       }
#if 0
       DEB(cout<<"\nPer system summary on maxratio with respect to the "
                 "maximum execution time:\n"
                 "[nCPU]\t\t[maxlat/exec]\t\t[maxlat/max_exec]"<<endl );
//#if 0
       Out.max<<"\nPer system summary on maxratio w.r.t. max_exec:\n"
                 "[nCPU]\t[maxlat/exec]\t[maxlat/max_exec]"<<endl ;
//#endif
       for(int i = minCPU - 1 ; i < M ; i++) {
           cout<<i+1<<"\t\t"
                   <<setw(11)
		   <<maxratio_per_CPU[i].maxlat/maxratio_per_CPU[i].exec_time
                   <<"\t\t"
		   <<setw(11)<<maxratio_per_CPU[i].maxlat/
		     maxratio_per_CPU[i].max_exec_time
		   <<endl ;
//#if 0
           Out.max<<i+1<<"\t"
                   <<maxratio_per_CPU[i].maxlat/maxratio_per_CPU[i].exec_time
                   <<"\t"
		   <<maxratio_per_CPU[i].maxlat/
		     maxratio_per_CPU[i].max_exec_time
		   <<endl ;
//#endif
       }
#endif
}

void reorder_decreasing(double vett[], int dim)
{
	for (int i = 0 ; i < dim ; i++) {
		int max_idx = i ; 
		for (int j = i + 1 ; j < dim ; j++) {
			if (vett[j] > vett[max_idx])
				max_idx = j ;
		}

		double tmp = vett[i] ;
		vett[i] = vett[max_idx] ;
		vett[max_idx]  = tmp ;
	}
}

int main(int argc, char* argv[])
{

  cout<<"\nSimulation of EDF scheduling of periodic tasks over a SMP.\n" ;

  /*
   * The default values for most parameters 
   * can be changed upon invocation.
   * The new value for dur possibly inserted from the command line is automatically 
   * multiplied by scale * maxNormPeriod.
   */
  if (argc > 1 && !strcmp(argv[1], "help")) {
    cout<<endl<<"Usage: "<<endl ;
    cout<<argv[0]<<" restart"<<endl ;
    cout<<"\tRead latest saved state from file "
	<<STATE_FNAME<<" in current dir and restart"<<endl ;
    cout<<endl ;
    cout<<argv[0]<<" Parameters"<<endl ;
    cout<<"Parameters:"<<endl<<"\t"<<
          "[minCPU="<<minCPU<<"] [maxCPU="<<maxCPU<<"] [nIter="<<nIter<<"]"<<
          " [Uminheavy="<<Uminheavy<<"] [Umaxheavy="<<Umaxheavy<<"]"<<
          " [heavyFraction="<<heavyFraction<<"]"<<
	  endl<<"\t"<<
          "[totUcapRatio="<<totUcapRatio<<"]"<<
          " [maxNormPeriod="<<maxNormPeriod<<"]"<<
          " [minNormPeriod="<<minNormPeriod<<"]"<<
          " [dur="<<dur<<"]"
	<<endl ;
    cout<<"\t-> dur is measured in number of maximum periods (i.e., the actual duration"
	<<endl
	<<"\t   will be computed by multiplying dur by "<<scale<<
          "*"<<maxNormPeriod<<")"
	<<endl ;
    exit(0) ;
  } 
  if (argc > 1 && !strcmp(argv[1], "restart")) { // restarting
    load_state() ;
  } else { // cold start: setup all paramenters
    DDB(cout<<"Cold start ..."<<endl) ;
    DDB(if (argc > 1) cout<<"Reading args ..."<<endl) ;

    if (argc > 1) 
      minCPU = atoi(argv[1]);
    else
      cout<<"Using default settings ...\n" ;
    M = minCPU ; // first system to simulate
    if (argc > 2) maxCPU = atoi(argv[2]);

    mindiff_per_CPU = new per_task_stat[maxCPU] ;
    maxdiff_per_CPU = new per_task_stat[maxCPU] ;
    minratio_per_CPU = new per_task_stat[maxCPU] ;
    maxratio_per_CPU = new per_task_stat[maxCPU] ;
    maxratio_UJ_per_CPU = new per_task_stat[maxCPU] ;
    maxlat_per_CPU = new per_task_stat[maxCPU] ;
    
    if (argc > 3) nIter = atoi(argv[3]);
    if (argc > 4) Umaxlight = Uminheavy = atof(argv[4]);
    if (argc > 5) Umaxheavy = atof(argv[5]);
    if (argc > 6) heavyFraction = atof(argv[6]);
    if (argc > 7) totUcapRatio = atof(argv[7]);
    if (argc > 8) maxNormPeriod = atoi(argv[8]) ;
    if (argc > 9) minNormPeriod = atoi(argv[9]) ;
    // read duration (to be then multiplied by scale * maxNormPeriod)
    if (argc > 10) dur = ((unsigned long long int) atoi(argv[10])) ;

    if (minNormPeriod > maxNormPeriod) {
      cout<<"minNormPeriod > maxNormPeriod"<<endl ;
      exit(1) ;
    }
    /* Reading a seed for the random numbers generator */
    ifstream rand_num("rand_gen.txt");
    DDB(cout<<"Opened rand_gen.txt"<<endl) ;
    if (rand_num.fail()) {
        cerr<<"Error in opening file rand_gen.txt\n" ;
        exit(-1) ;
    }
    rand_num >> seed;
    rand_num.close();
    srand(seed);
  } // end of setting of all parameters in case of cold start

  // some sanity checks
  if (heavyFraction < 0 || heavyFraction > 1) {
      cout<<"Ouch! Heavy fraction not in [0, 1]"<<endl ;
      exit(1) ;
  }

  if (worstCaseScenario && heavyFraction > 0 &&
	totUcapRatio * heavyFraction <
	(Umaxheavy * (maxCPU - 1))/maxCPU - MIN_DOUBLE_CONST) {
	  heavyFraction = (Umaxheavy * (maxCPU - 1)) / maxCPU / totUcapRatio ;

	  cout<<"Worst-case-scenario mode set and non-null heavy fraction,"
	      <<endl<<"but the latter is not high enough to guarantee room for"
	      <<endl<<"M - 1 tasks with maximum possible utilization"
	      <<endl<<"(if heavy fraction is not null, then the first M - 1"
	      <<endl<<"tasks will be heavy)"
	      <<endl<<endl<<"Automatically raising heavyFraction to "
	      <<heavyFraction
	      <<endl ;
  }

  if (worstCaseScenario && heavyFraction > 0 &&
	totUcapRatio * heavyFraction <
	(Umaxheavy * (maxCPU - 1))/maxCPU - MIN_DOUBLE_CONST) {
	  cout<<"Wrong update of heavyFraction"<<endl ;
	  exit(-1) ;
  }

  if (Uminheavy == 0)
	Uminheavy = (double) Umin_rep / scale ;

  if (Uminheavy <= 0 || Uminheavy > 1) {
    cout<<"Ouch! Uminheavy "<<Uminheavy
	<<" not in (0, 1]"<<endl ;
    exit(1) ;
  }
  // set actual duration, capped through maxSupportedDur
  dur *= scale * maxNormPeriod ;
  if (dur > maxSupportedDur)
	dur = maxSupportedDur ;
  /* 
   * From the considerations about he maximum duration
   * supported by rtsim, it follows that the maximum 
   * normalized period usch that all tasks are executed
   * at least once is as follows.
   */
  const int maxSupportedPeriod = dur/scale ;
  if (maxNormPeriod > maxSupportedPeriod) {
    cout<<"maxNormPeriod "<<maxNormPeriod
	<<" > maxSupportedPeriod "<<maxSupportedPeriod
	<<endl ;
    exit(1) ;
  }

  int Uminheavy_rep = (int) (Uminheavy * scale) ;
  if (Uminheavy_rep < Umin_rep)
	Uminheavy_rep = Umin_rep ;

  const int Umaxheavy_rep = (int) (Umaxheavy * scale) ;

  // We need to estimate the maximum number of tasks for each task set.
  // compute maximum target total utilization
  const int maxTargetUtot_rep = (int) (totUcapRatio * maxCPU  * scale) ;
  DDB(cout<<"maxTargetUtot_rep "<<maxTargetUtot_rep
	  <<" = "<<totUcapRatio
	  <<" * "<<maxCPU
	  <<" * "<<scale
	  <<endl) ;
  /* 
   * Compute the maximum possible number of heavy tasks as a function 
   * of their minimum utilization Uminheavy, and so as to let them use a 
   * fraction heavyFraction of the target total bandwidth.
   */
  const int maxNumHeavy = 
            (int)((maxTargetUtot_rep * heavyFraction ) / Uminheavy_rep + 1) ;

  // estimate an upper bound maxlight for the number of light tasks
  const int maxNumLight = (int)(2 *( lightRatio * 2 * maxTargetUtot_rep / 
                            (Umin_rep + Uminheavy_rep - 1) )) ;

  const int maxNumTask = maxNumHeavy + maxNumLight ;
  Tick period[maxNumTask];

  // integer representation of task utilizations
  int U_rep[maxNumTask];

  // task descriptors
  RTTask *tt[maxNumTask] ;

  // per-task maximum and mean tardiness observers
  maxLat *maxL[maxNumTask];
  meanLat *meanL[maxNumTask];
  cerr.precision(5) ;
  cout.precision(5) ;

  try {

#if 0
    Simulation::dbg.setStream("log.txt");
    Simulation::dbg.enable(_SERVER_DBG_LEV);
    Simulation::dbg.enable(_TASK_DBG_LEV);
    Simulation::dbg.enable(_KERNEL_DBG_LEV);
    Simulation::dbg.enable(_INSTR_DBG_LEV);
    Simulation::dbg.enable(_GRUB_DBG_LEV);

    JavaTrace jtrace("trace.trc");
    Out_t Out ;
    Out.mean.open("mean_tard.dat",ofstream::out);
    Out.max.open("max_tard.dat",ofstream::out);
#endif

    // print configuration
    cout<<"Configuration:"<<endl ;
    cout<<"Random seed: "<<seed<<endl ;
    cout<<"minCPU: "<<minCPU<<", maxCPU: "<<maxCPU
	<<", total number of iterations: "<<nIter<<endl ;
    if (M != minCPU || iter != 1) {
      cout<<"Warm restart: " ;
      if (iter > nIter) {
	M++ ;
	iter = 1 ;
      }
      if (M > maxCPU) {
	cout<<"last run completed all simulations, "
	    <<"just print statistics now"<<endl ;
      } else
	cout<<" restarting from "<<M<<" CPUs, iteration "<<iter<<endl ;
    }
    else
      cout<<"Cold start"<<endl ;
    cout<<"Task utilizations and periods are multiples of, "
          "respectively, 1/"<<scale<<" and "<<scale<<endl ;
    cout<<"Minimum task utilization: "
        <<(double) Umin_rep/scale<<endl ;
    cout<<"Minimum heavy task utilization: "
        <<(double) Uminheavy_rep/scale<<endl ;
    cout<<"Maximum heavy task utilization: "<<Umaxheavy<<endl ;
    cout<<"Maximum task period: "<<maxNormPeriod<<" * "<<scale<<" ticks"<<endl ;
    cout<<"Minimum task period: "<<minNormPeriod<<" * "<<scale<<" ticks"<<endl ;
    cout<<"Worst-case-scenario mode "<<worstCaseScenario<<endl ;
    cout<<"Length of each simulation: "<<dur/(scale*maxNormPeriod)<<" * "
        <<maxNormPeriod<<" * "<<scale<<" ticks"<<endl ;
    cout<<"Ratio between target total utilization and system capacity: "
        <<totUcapRatio<<endl ;
    cout<<"Maximum fraction of target total utilization assigned to "
	<<"heavy tasks: "<<heavyFraction<<endl ;
    cout<<"Ratio between actual number of light tasks"
          " and minimum number needed: "<<lightRatio<<endl ;
   

    // the initial value of M has already been properly set in case of both
    // cold and warm start
    for ( ; M <= maxCPU ; M++) { // system made of M CPUs
        const int m = M - 1 ;
        // compute target total utilization
        const double targetUtot = totUcapRatio * M ;
        // integer representation, ranging from 0 to M * scale
        const int targetUtot_rep = (int) (targetUtot * scale) ;
  	DDB(cout<<"targetUtot_rep "<<maxTargetUtot_rep
	  <<" = "<<totUcapRatio
	  <<" * "<<M
	  <<" * "<<scale
	  <<endl) ;
	DDB(cout<<"targetUtot "<<targetUtot<<", targetUtot_rep "<<targetUtot_rep<<endl) ;
        if (targetUtot_rep > maxTargetUtot_rep) {
		cout<<"Target Utot rep inconsistent with respect to max "
		    <<maxTargetUtot_rep<<endl ;
		exit(-1) ;
	}
        // max fraction of the target total utilization destined to heavy tasks
        const double targetUtotHeavy = heavyFraction * targetUtot ;
        // integer representation, ranging from 0 to M * scale
        int targetUtotHeavy_rep = (int) (targetUtotHeavy * scale) ;
	DDB(cout<<"targetUtotHeavy "<<targetUtotHeavy
		<<" = "<<heavyFraction<<" * "<<targetUtot
		<<", targetUtotHeavy_rep "<<targetUtotHeavy_rep<<endl) ;

        if (targetUtotHeavy_rep > targetUtot_rep) {
		cout<<"Target UtotHeavy_rep inconsistent"<<endl ;
		if (targetUtotHeavy_rep <= targetUtot_rep + 1) {
			cout<<"But very close, adjusting ..."<<endl ;
			targetUtotHeavy_rep = targetUtot_rep ;
		} else
			exit(-1) ;
	}

        if (iter == 1) { // first iteration for this system 
	  // init statistics for this system
	  mindiff_per_CPU[m].diff = -MIN_DOUBLE_CONST - 1 ; 
	  maxdiff_per_CPU[m].diff = 0 ;
	  minratio_per_CPU[m].ratio = -1 ;
	  maxratio_per_CPU[m].ratio = 0 ;
	  maxratio_UJ_per_CPU[m].ratio_UJ = 0 ;
	  maxlat_per_CPU[m].maxlat = 0 ;
	} // otherwise all above data structure are alredy properly initialized

        cout<<"\n------------------------------------------------------"<<endl ;
        cout << "**** Simulating " << M << "/"<<maxCPU<<" CPUs - " ;
        cout<< " target total utilization: "<<targetUtot<<endl ;
#if 0
        Out.mean << "M = " << M << ":"<<endl ;
        Out.max << "M = " << M << ":"<<endl ;
#endif

	// the initial value can be either 1, 
	// or larger than 1 in case of a restart
        for ( ; iter <= nIter ; iter ++) {
            EDFScheduler sched;
            /* Kernel with more internal CPUs (NOT capable of power saving) */
            MRTKernel edfkern(&sched,M);

            cout<< "\nIteration "<<iter<<"/"<<nIter<<" ("<<M<<"/"<<maxCPU<<" CPUs): " ;
#if 0
            Out.mean<< "\tIteration "<<iter<<":"<<endl ;
            Out.max<< "\tIteration "<<iter<<":"<<endl ;
#endif
            // integer representation of current total utilization
            int totU_rep = 0;
            int index = 0 ;

	    // reduce the probability to generate again a previous task set
	    srand(labs(rand() + iter)) ;
	    
            // generate utilizations of heavy tasks
            while(totU_rep < targetUtotHeavy_rep) {
		    U_rep[index] = Uminheavy_rep + 
                                rand()%(Umaxheavy_rep - Uminheavy_rep + 1) ;

                DDB(cout<<"\nUtilization of heavy task "<<index<<": "
                    <<(double) U_rep[index]/scale<<endl ;)

                if (totU_rep + U_rep[index] > targetUtotHeavy_rep){
                   // cap utilization: this is certainly the last loop for
                   // this phase
		   
                   U_rep[index] = targetUtotHeavy_rep - totU_rep ;
		   DDB(cout<<"Gone beyond limit, task "
		       <<index<<" tentative U "
		       <<static_cast<double>(U_rep[index])/scale
		       <<endl) ;

		   if (U_rep[index] < Umin_rep) { // invalidate U_rep[index]
			DDB(cout<<"Tentative U too small: "<<U_rep[index]<<" "<<Umin_rep<<endl) ;
			break ; // we are now as close as possible
				// to targetUtotheavy
		   }
                }
                totU_rep += U_rep[index];
                
                index++ ;// add a new heavy task and work on next task
            } // end of bandwidth assignment to heavy tasks
      	    const int nheavy = index ;

	    if (worstCaseScenario && nheavy > 0 && nheavy < ceil(static_cast<double>(M - 1)/2)) {
		    cout<<"Worst-case-scenario mode set with heavy tasks,"
			<<"but nheavy < M - 1"
			<<endl ;
		    exit(-1) ;
	    }

#if 0 // most certainly not needed any more
            // Expansion loop: if needed, increase heavy tasks'
	    // utilizations until totU = targetUtotHeavy; however, abort
	    // if there is room for some light tasks
            index = 0 ;
            for (int avail = targetUtotHeavy_rep - totU_rep ; 
               avail > 0 && avail > Uminheavy_rep &&
			 index < nheavy ; index ++) {
              const int feasible = avail < Umaxheavy_rep - U_rep[index] ? 
                                  avail : Umaxheavy_rep - U_rep[index] ;
              U_rep[index] += feasible ;
	      DDB(cout<<"\nNew utilization of heavy task "<<index<<": "
		  <<(double) U_rep[index]/scale<<endl ;)

              totU_rep += feasible ;
              avail -= feasible ;
            } // end of expansion loop
#endif
            
      	    // correcteness checks on heavy tasks total utilization
	    if (totU_rep < targetUtotHeavy_rep - Umin_rep) {
		    cout<<"totUrep "<<totU_rep
			<<" targetUtotHeavy_rep "
			<<targetUtotHeavy_rep
			<<endl ;
		    exit(-1) ;
	    }
	    int tot_check_heavy = 0 ;	  
	    for (int i = 0 ; i < nheavy ; i++) 
	   	tot_check_heavy += U_rep[i] ;
	    if (tot_check_heavy != totU_rep) {
		cout<<"\nWhat's happening here ?"<<endl ;
	        cout<<"Heavy tasks' total utilization: "
		    <<tot_check_heavy<<
		  " Incremental total utilization: "<<totU_rep<<endl ;
	  	exit(-1) ;
	    }

       	    cout<<" "<<nheavy<<" heavy tasks (tot. utilization "
                <<(double) totU_rep/scale<<")" ;
            if (nheavy > maxNumHeavy) {
                cout<<"\nHey, spendthrift! You forecast only "<<maxNumHeavy<<
                      " heavy tasks, but you generated "<<nheavy<<" of them\n" ;
                exit(-1) ;
            }

            const int Umaxlight_rep = Umaxlight * scale ;

            // compute an upper bound to the number of light tasks
            const double fp_light = lightRatio * 2 *
                                    (double) (targetUtot_rep - totU_rep) / 
                                    (Umin_rep + Umaxlight_rep - 1) ;

            DDB(cout<<"\nfp_light "<<fp_light<<endl ;)
            
            // round up to get the number of light tasks to generate
            int nlight ;
	    if (targetUtot_rep - totU_rep >= Umin_rep)
		nlight = ((int) fp_light < fp_light ? 
                               fp_light + 1 : fp_light );
	    else
		nlight = 0 ;

            if (nlight > maxNumLight) {
                cout<<"\nHey, spendthrift! You forecast only "<<maxNumLight<<
                      " light tasks, but you generated "
		    <<nlight<<" of them\n" ;
                exit(-1) ;
            }
            int ntask = nheavy + nlight ; // may be reduced below

	    if (nlight > 1) { // fill remaining bandwidth with light tasks
#if 0
              // compute Umeanlight
              const int Umeanlight_rep = (targetUtot_rep-totU_rep)/nlight ;
	      DDB(cout<<"Umeanlight_rep: "<<Umeanlight_rep<<endl) ;
	      DDB(cout<<"Umin_rep: "<<Umin_rep<<endl) ;
	      
              // compute more appropriate Umaxlight
              const int Umaxlight_rep = 2*Umeanlight_rep - Umin_rep ;
              if (Umaxlight_rep <= Umin_rep) {
                cout<<"You, cheapskate! Umaxlight must be at least "<<Umin_rep<<
                      " and you generated "<<Umaxlight_rep<<endl;
                exit(-1) ;
              }
#endif
              if (Umaxlight_rep > Uminheavy_rep) {
                cout<<"Hey, spendthrift! Umaxlight must be at most "
                    <<Uminheavy_rep-1<<" and you generated "
                    <<Umaxlight_rep<<endl;
                exit(-1) ;
              }

              // randomly generate the utilizations of light tasks
              for (index = nheavy ; index < ntask ; index++) {
                    U_rep[index] = Umin_rep +
			    rand()%(Umaxlight_rep - Umin_rep + 1);

                  totU_rep += U_rep[index];
                  if (totU_rep >= targetUtot_rep)
                      break ;
              } // end of bandwidth assignment to light tasks

              DDB(cout<<"After first bandwidth assignment: "
                    <<(double) totU_rep/scale<<endl) ;
            
              if (index < ntask) {
                DDB(cout<<"Shrinking needed"<<endl) ;
                // sync nlight and ntask with the actual number of tasks
                nlight -= ntask - index - 1 ;
                ntask = index + 1 ;

                // try to reduce utilization of the last task to fit targetUtot
                int new_U_rep = U_rep[index] - (totU_rep - targetUtot_rep) ;

		if (new_U_rep < Umin_rep) { // remove also last task
			nlight-- ;
			ntask-- ;
			totU_rep -= U_rep[index] ;
		} else {
			U_rep[index] = new_U_rep ;
                	totU_rep = targetUtot_rep ;
		}
              } // end of shrinking

              DDB(if (totU_rep < targetUtot_rep)
		    cout<<"Expansion needed"<<endl) ;

              // if needed, increase light task util. until totU=targetUtot
              index = nheavy ;
              for (int avail = targetUtot_rep - totU_rep ; 
                 avail > 0 && index < ntask ; index ++) {
                const int feasible = avail < Umaxlight_rep - U_rep[index] ? 
                                    avail : Umaxlight_rep - U_rep[index] ;
                U_rep[index] += feasible ;
                totU_rep += feasible ;
                avail -= feasible ;
              } // end of expansion loop
            } // end of case nlight > 1
	    else if (nlight == 1) { // just assign the spare bandwidth to it
		  if (targetUtot_rep - totU_rep > Uminheavy_rep-1) {
                    cout<<"There's something wrong! nlight == 1, but spare "
		    	<<"bandwidth is not lower than the minimum utilization "
			<<(double) Uminheavy_rep/scale<<" for a heavy task"
			<<endl ;
                	exit(-1) ;
              	  }
		  if (targetUtot_rep - totU_rep < Umin_rep) {
                    cout<<"There's something wrong! nlight == 1 and spare "
		    	<<"bandwidth is lower than the minimum utilization "
			<<(double) Umin_rep/scale<<endl ;  
                	exit(-1) ;
              	  }
		  DDB(cout<<"\nUtilization of single light task "<<nheavy<<": "
		      <<(double) U_rep[nheavy]/scale<<endl ;)

		  U_rep[nheavy] = targetUtot_rep - totU_rep ;
		  totU_rep += U_rep[nheavy] ;
	    } // end of case nlight == 1

            cout<<", "<<nlight<<" light tasks"<<endl ;

            if (totU_rep < targetUtot_rep - Umin_rep ||
		totU_rep > targetUtot_rep) {
                cout<<"Ouch! After all such labour, I got "
                    <<(double) totU_rep/scale
                    <<" instead of the target total utilization "<<
                      targetUtot
		    <<"i,e., totU_rep "<<totU_rep
		    <<", instead of targetUtot_rep "<<targetUtot_rep
		    <<endl ;
                exit(-1) ;
            }

	    // correctenss check on total utilization
	    int tot_check = 0 ;	  
	    for (int i = 0 ; i < ntask ; i++) 
	   	tot_check += U_rep[i] ;
	    if (tot_check > targetUtot_rep) {
		cout<<"What the hell are you doing?"<<endl ;
	        cout<<"Scaled Target utilization: "
		    <<(double)targetUtot_rep<<
		  " Incremental total utilization: "<<totU_rep<<
	          " Actual utilization: "<<tot_check<<endl ;
	  	exit(-1) ;
	    }

     	    DDB(cout<<"Final total utilization: "
                    <<(double) totU_rep/scale<<endl) ;

            // find minimum and maximum utilization in the task set (used later)
            int maxUindex = 0, minUindex = 0;
            for (int i = 1 ; i < ntask ; i++) {
               if (U_rep[i] > U_rep[maxUindex])
                   maxUindex = i ;
               else
                   minUindex = U_rep[minUindex] < U_rep[i] ? minUindex : i ;
            }
	    
            try {
                  // generate tasks and attach observers
	    	  int exec_time[ntask] ;
	    	  int max_exec_time = 0 ; // maximum task execution time 
		  int min_exec_time = -1 ; // minimum task execution time
                  for (int i = 0; i < ntask; i++) {
                     // first generate period
		     if (!worstCaseScenario)
			     period[i] = ( minNormPeriod + rand()%maxNormPeriod ) * scale ;
		     else  {
			if (i == 0 || i >= nheavy) // first heavy task or light task -> large period
			     period[i] = ( (maxNormPeriod*4)/3 + 1 + rand()%((maxNormPeriod*4)/3) ) * scale ;
			else // any other heavy task -> small period
			     period[i] = ( minNormPeriod + rand()%(maxNormPeriod/10)) * scale ;
		    }

                     char nomeTask[100] ;
                     sprintf(nomeTask, "Task %d", i) ;

		     cout<<((i < nheavy) ? "Heavy " : "Light ") ;
	             cout << "task " << i << ": " ;
                     DEB(cout<<"utilization "
                             <<(double) U_rep[i]/scale << ", " ) ;

		     DDB(cout<<"mininterarrival "<<period[i]
			 <<"maxinterarrival "
			 <<period[i] * (1 + sporad_coeff)
			 <<endl) ;
                     tt[i] = 
                         new PeriodicTask(period[i], period[i], 0, nomeTask);
// it seems that the following sporadic pattern yields lower lateness
//			     new RTTask(new UniformVar(period[i],
//						       period[i] *
//						       (1 + sporad_coeff)),
//					period[i], 0, nomeTask);
		     
                     tt[i]->setAbort(false);
#if 0
		     tt[i]->setTrace(&jtrace);
#endif
		     // E = U * P = (U_rep/scale) * P = U_rep * (P/scale)
                     exec_time[i] = U_rep[i] * ((int) period[i]/scale) ;

                     DEB(cout<<"execution time "<<exec_time[i] ;
                         cout<<", period "<<period[i]<<endl ) ;

		     // check on utilization
		     if (U_rep[i] != 
                         ((double) exec_time[i]/((double) period[i]/scale))) {
			double exec = exec_time[i], per = period[i] ;
			double util = exec/per ;
  			cout.precision(50) ;
			cout<<"\nEhm... U_rep/scale = "
			    <<(double) U_rep[i]/scale
                            <<" exec_time/period = "
                            << util << " difference = "
			    <<(double) U_rep[i]/scale - util
                            <<" difference * scale = "
                            << U_rep[i] - exec/(per/scale)
			    <<endl ;
			exit(-1) ;
		     }
                     max_exec_time = max_exec_time > exec_time[i] 
		                     ? 
                                     max_exec_time : exec_time[i] ;
		     min_exec_time = min_exec_time < exec_time[i] &&
		                     min_exec_time >= 0 
		                     ? 
		                     min_exec_time : exec_time[i] ;
                     char digits[100] ;
                     sprintf(digits, "%d", exec_time[i]) ;
                     string s = string("delay(") + 
                                string(digits) + string(");");
                     tt[i]->discardInstrs();
                     tt[i]->insertCode(s);
                     sched.addTask(tt[i]);

                     string tname1 = "";
                     string tname2 = "";
/*
                     if (i == minUindex) {
                         tname1 = string("Max_Tardiness_for_minU");
                         tname2 = string("Mean_Tardiness_for_minU");
                     } else if (i == maxUindex) {
                         tname1 = string("Max_Tardiness_for_maxU");
                         tname2 = string("Mean_Tardiness_for_maxU");
                     }
*/
                     maxL[i] = new maxLat(tname1);
                     meanL[i] = new meanLat(tname2);
                     
                     maxL[i]->attach(tt[i]);
                     meanL[i]->attach(tt[i]);
                  } // end of assignment of task parameters and watches

       		  // corectness check on total utilization (from task param.)
		  double tot_check = 0 ;	  
		  for (int i = 0 ; i < ntask ; i++) {
		  	tot_check += (double) exec_time[i]/period[i] ;
		  } 
		  if (tot_check > M + MIN_DOUBLE_CONST) {
			cout<<"Total utilization "<<tot_check
			    <<" higher than "<<M
			    <<endl ;
			exit(-1) ;
		  }
		  if( tot_check < M * totUcapRatio - UTIL_PRECISION
                      ||
		      tot_check > M * totUcapRatio + UTIL_PRECISION) {
			cout<<"What the hell are you doing?"<<endl ;
  			cout.precision(50) ;
			cout<<"Target utilization: "<<M*totUcapRatio<<
			      " Actual utilization: "<<tot_check<<
			      " Difference: "<<tot_check - M*totUcapRatio<<
                              endl 
			      <<"ntask "<<ntask<<endl<<endl ;
		  	for (int i = 0 ; i < ntask ; i++) {
			    double exec = exec_time[i], per = period[i] ;
		  	    cout<<"Task "<<i<<"util. "<<exec/per<<endl ;
		  	} 
		  	exit(-1) ;
		  }
                  DEB(cout<<"Maximum execution time: "<<max_exec_time<<endl) ;

                  // perform simulation
                  Simulation::run(dur, 1);
	
                  cout<<endl ;

                  // Compute and log statistics for this iteration.
                  // Especially, aggregated statistics for this iteration are:
                  // maxdiff, mindiff, maxratio and minratio
                  per_task_stat mindiff = { 
                                           0,                    // task
                                           0,                    // iter
                                           0,                    // CPUs
					   0,			 // exec time
					   0,			 // max exec
                                           0,                    // th. bound
                                           0,                    // th. b. UJ
                                           0,                    // max lat
                                           0,                    // mean lat
                                           -MIN_DOUBLE_CONST -1, // diff
                                           0,                    // ratio
                                           0                     // ratio_UJ
                                          } ;
                  per_task_stat maxdiff = 
		    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;
                  per_task_stat minratio = 
		    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0 } ;
                  per_task_stat maxratio = 
		    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;
                  per_task_stat maxratio_UJ = 
		    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
                  per_task_stat maxlat = 
		    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } ;

#if 0
		  // array of exec times in decreasing order
		  double decreasing_exec[ntask] ;
		  // array of utilizations in decreasing order
		  double decreasing_U[ntask] ;
		  // array of M_g in decreasing order
		  double decreasing_M[M] ;
		  // array of betas in decreasing order
		  double decreasing_beta[ntask] ;

		  for (int i = 0 ; i < ntask; i++) {
			  decreasing_exec[i] = exec_time[i] ;
			  decreasing_U[i] = (double) U_rep[i]/scale ;
		  }
		  
		  reorder_decreasing(decreasing_exec, ntask) ;
		  reorder_decreasing(decreasing_U, ntask) ;

		  cout<<"Exec:"<<endl ;
		  for (int i = 0 ; i < ntask ; i++)
			  cout<<decreasing_exec[i]<<endl ;

		  cout<<"U:"<<endl ;
		  for (int i = 0 ; i < ntask ; i++)
			  cout<<decreasing_U[i]<<endl ;

		  decreasing_M[0] = M ;
		  for (int i = 1 ; i < M ; i++) {
			  cout<<"Formula for M["<<i+1<<"]: "
			      <<decreasing_M[i-1]<<" - "
			      <<decreasing_U[i-1]
			      <<endl ;
			  decreasing_M[i] = decreasing_M[i-1] -
				  decreasing_U[i-1];
		  }

		  cout<<"Mg:"<<endl ;
		  for (int i = 0 ; i < M ; i++)
			  cout<<decreasing_M[i]<<endl ;

		  double base_lambda = 0 ;
		  for (int i = 0 ; i < M - 1 ; i++) {
			  cout<<"Formula for term "<<i+1<<": "
			      <<decreasing_exec[M-2-i]<<" / "
			      <<decreasing_M[i]
			      <<endl ;
			  base_lambda += decreasing_exec[M-2-i] /
				  decreasing_M[i] ;
		  }

		  base_lambda /= decreasing_M[M-1] ;

		  cout<<"Base lambda: "<<base_lambda<<endl ;

		  for (int i = 0 ; i < ntask; i++) {
			  double U_i = (double) U_rep[i]/scale ;

			  decreasing_beta[i] =
				  U_i * base_lambda +
				  ((double) exec_time[i]) * (M-U_i) / M ;
		  }
		  reorder_decreasing(decreasing_beta, ntask) ;
		  
		  cout<<"Betas:"<<endl ;
		  for (int i = 0 ; i < ntask ; i++)
			  cout<<decreasing_beta[i]<<endl ;

                  double th_max_lat_base = 0 ;

		  for (int i = 0 ; i < M - 1 ; i++)
			  th_max_lat_base += decreasing_beta[i] ;
		  th_max_lat_base /= M ;
		  
		  const double M_1M = (double) (M-1)/M ;

		  // computing base part of (36) in Uma-Anderson bound
		  double totU = (double) totU_rep/scale ;
		  
		  int Lambda_UJ_bound = floor(totU) ;
		  double int_part ;
		  if (modf(totU, &int_part ) <
		      std::numeric_limits<double>::epsilon()) // is integral
			  Lambda_UJ_bound-- ;

                  double th_max_lat_base_UJ = -min_exec_time ;

		  for (int i = 0 ; i < Lambda_UJ_bound ; i++)
			  th_max_lat_base_UJ += decreasing_exec[i] ;
		  
		  th_max_lat_base_UJ /= decreasing_M[Lambda_UJ_bound-1] ;
                  DDB(cout<<"Base th. upper bound: "<<th_max_lat_base<<endl) ;
#endif

                  for (int i = 0; i < ntask; i++) {
                      per_task_stat cur_task ;
                      cur_task.task = i ;
                      cur_task.CPU = M ;
                      cur_task.iter = iter ;
		      cur_task.exec_time = exec_time[i] ;
		      cur_task.max_exec_time = max_exec_time ;

#if 0
                      cur_task.th_bound = th_max_lat_base +
			      M_1M * exec_time[i] ;

                      cur_task.th_bound_UJ = th_max_lat_base_UJ +
			      exec_time[i] ;
#else
                      cur_task.th_bound = -1 ;

                      cur_task.th_bound_UJ = -1 ;
#endif
                      cur_task.maxlat = maxL[i]->getMean() ;
                      cur_task.meanlat = meanL[i]->getMean() ;

#if 0
                      cur_task.diff = cur_task.th_bound - cur_task.maxlat ;
                      cur_task.ratio = cur_task.maxlat / cur_task.th_bound ;
                      cur_task.ratio_UJ = cur_task.maxlat / 
			                  cur_task.th_bound_UJ ;
#endif

                      DEB(cout<<"Task "<<cur_task.task<<": mean lateness "
                              <<cur_task.meanlat<<", max lateness "
                              <<cur_task.maxlat<<", th. bound None"
#if 0
                              <<cur_task.th_bound<<", th. bound UJ "
                              <<cur_task.th_bound_UJ<<", diff "
                              <<cur_task.diff<<", ratio "
                              <<cur_task.ratio<<", ratio_UJ "
                              <<cur_task.ratio_UJ
#endif
			      <<endl ) ;

#if 0
                      Out.mean<<"\t\t"<<cur_task.meanlat<<endl ;
                      Out.max<<"\t\t"<<cur_task.maxlat<<"\t"
                             <<cur_task.th_bound<<"\t"<<cur_task.diff
                             <<"\t"<<cur_task.ratio<<endl ;

                      if (cur_task.diff < -MIN_DOUBLE_CONST) {
                          cout<<"\nOuch!!!!!! Broken bound:"<<endl ;
                          cout<<M<<" CPUs, iteration "<<iter<<endl ;
                          cout<<"Task "<<i<<": "
                              <<"actual maximum lateness "<<cur_task.maxlat
                              <<", theoretical bound "<<cur_task.th_bound
                              <<endl<<endl ;

	                  jtrace.close();
                          exit(-1) ;
                      }
#endif

                      // update aggregated statistics for this iteration, i.e.
                      // maxdiff, mindiff, maxratio and minratio
#if 0
                      if (cur_task.diff > maxdiff.diff) 
                          maxdiff = cur_task ;
                      
                      if (mindiff.diff < -MIN_DOUBLE_CONST || 
                          cur_task.diff < mindiff.diff)
                          mindiff = cur_task ;
                          
                      if (cur_task.ratio > maxratio.ratio) 
                          maxratio = cur_task ;

                      if (cur_task.ratio_UJ > maxratio_UJ.ratio_UJ) 
                          maxratio_UJ = cur_task ;
                          
                      if (minratio.ratio < 0 || 
                          cur_task.ratio < minratio.ratio) 
                          minratio = cur_task ;
#endif
                      if (cur_task.maxlat > maxlat.maxlat) 
                          maxlat = cur_task ;

                  } // end of statistic loop

                  // log aggregated statistics for this iteration, i.e.
                  // maxdiff, mindiff, maxratio and minratio
#if 0
                  cout<<"Minimum diff "<<mindiff.diff<<" for task "
		      <<mindiff.task<<endl ;
                  cout<<"Maximum diff "<<maxdiff.diff<<" for task "
		      <<maxdiff.task<<endl ;
                  cout<<"Minimum ratio "<<minratio.ratio<<" for task "
		      <<minratio.task<<endl ;
                  cout<<"Maximum ratio "<<maxratio.ratio<<" for task "
		      <<maxratio.task<<endl ;
                  cout<<"Maximum ratio UJ"<<maxratio_UJ.ratio_UJ<<" for task "
		      <<maxratio_UJ.task<<endl ;
#endif
                  cout<<"Maximum lateness "<<maxlat.maxlat<<" for task "
		      <<maxlat.task<<endl ;
#if 0
                  Out.max<<"\t\tMin_diff "<<mindiff.diff<<" task "
                         <<mindiff.task<<endl ;
                  Out.max<<"\t\tMax_diff "<<maxdiff.diff<<" task "
                         <<maxdiff.task<<endl ;
                  Out.max<<"\t\tMin_ratio "<<minratio.ratio<<" task "
                         <<minratio.task<<endl ;
                  Out.max<<"\t\tMax_ratio "<<maxratio.ratio<<" task "
                         <<maxratio.task<<endl ;
                  Out.max<<"\t\tMax_lateness "<<maxlat.maxlat<<" task "
                         <<maxlat.task<<endl ;
#endif
#if 0
                  // update aggregated statistics for this system, i.e. 
                  // mindiff_per_CPU[m], maxdiff_per_CPU[m], 
                  // minratio_per_CPU[m], maxratio_per_CPU[m]
                  if (maxdiff.diff > maxdiff_per_CPU[m].diff)
                      maxdiff_per_CPU[m] = maxdiff ;
                      
                  if (mindiff_per_CPU[m].diff < -MIN_DOUBLE_CONST || 
                           mindiff.diff < mindiff_per_CPU[m].diff) 
                      mindiff_per_CPU[m] = mindiff ;
                      
                  if (maxratio.ratio > maxratio_per_CPU[m].ratio) 
                      maxratio_per_CPU[m] = maxratio ;

                  if (maxratio_UJ.ratio_UJ > maxratio_UJ_per_CPU[m].ratio_UJ) 
                      maxratio_UJ_per_CPU[m] = maxratio_UJ ;
                      
                  if (minratio_per_CPU[m].ratio < 0 || 
                           minratio.ratio < minratio_per_CPU[m].ratio) 
                      minratio_per_CPU[m] = minratio ;
#endif

                  if (maxlat.maxlat > maxlat_per_CPU[m].maxlat) 
                      maxlat_per_CPU[m] = maxlat ;
               } // end of try block
               catch (ParseExc &pe) {
                   cout << pe.what() << endl;
                   exit(-1);
               }
	       // free memory
               for (int i = 0; i < ntask; i++) {
                   delete tt[i]; delete maxL[i]; delete meanL[i];
               }
               cout<<"---------------------------\n" ;
       	       // partial per system summaries
               cout<<"Partial summary after "<<iter<<"/"<<nIter
		   <<" iterations ("<<M<<" CPUs):" ;
       	       per_system_summaries(
#if 0
Out
#endif
) ;
	       save_state() ;
               cout<<"---------------------------\n" ;
           } // end of internal iterations loop

           // log aggregated statistics for this system, i.e.
           // maxdiff_per_CPU, mindiff_per_CPU, maxratio_per_CPU and 
           // minratio_per_CPU
           DEB(cout<<"Aggregated statistics with "<<M<<" CPUs:"<<endl ;
#if 0
               cout<<"\tMinimum diff "<<mindiff_per_CPU[m].diff
                   <<", occurred at iteration "<<mindiff_per_CPU[m].iter
                   <<", for task "<<mindiff_per_CPU[m].task<<endl ;
               cout<<"\tMax diff "<<maxdiff_per_CPU[m].diff
                   <<", occurred at iteration "<<maxdiff_per_CPU[m].iter
                   <<" for task "<<maxdiff_per_CPU[m].task<<endl ;
               cout<<"\tMinimum ratio "<<minratio_per_CPU[m].ratio
                   <<", occurred at iteration "<<minratio_per_CPU[m].iter
                   <<", for task "<<minratio_per_CPU[m].task<<endl ;
               cout<<"\tMax ratio "<<maxratio_per_CPU[m].ratio
                   <<", occurred at iteration "<<maxratio_per_CPU[m].iter
                   <<", for task "<<maxratio_per_CPU[m].task<<endl  ;
               cout<<"\tMax ratio UJ "<<maxratio_UJ_per_CPU[m].ratio_UJ
                   <<", occurred at iteration "<<maxratio_UJ_per_CPU[m].iter
                   <<", for task "<<maxratio_UJ_per_CPU[m].task<<endl  ;
#endif
               cout<<"\tMax lateness "<<maxlat_per_CPU[m].maxlat
                   <<", occurred at iteration "<<maxlat_per_CPU[m].iter
	           <<", for task "<<maxlat_per_CPU[m].task<<endl ;)
#if 0
           Out.max<<"\tMin_diff_"<<M<<"_CPUs "<<mindiff_per_CPU[m].diff
                  <<" iteration "<<mindiff_per_CPU[m].iter
                  <<" task "<<mindiff_per_CPU[m].task<<endl ;
           Out.max<<"\tMax_diff_"<<M<<"_CPUs "<<maxdiff_per_CPU[m].diff
                  <<" iteration "<<maxdiff_per_CPU[m].iter
                  <<" task "<<maxdiff_per_CPU[m].task<<endl ;
           Out.max<<"\tMin_ratio_"<<M<<"_CPUs "<<minratio_per_CPU[m].ratio
                  <<" iteration "<<minratio_per_CPU[m].iter
                  <<" task "<<minratio_per_CPU[m].task<<endl ;
           Out.max<<"\tMax_ratio_"<<M<<"_CPUs "<<maxratio_per_CPU[m].ratio
                  <<" iteration "<<maxratio_per_CPU[m].iter
                  <<" task "<<maxratio_per_CPU[m].task<<endl ;
           Out.max<<"\tMax_lateness_"<<M<<"_CPUs "<<maxlat_per_CPU[m].ratio
                  <<" iteration "<<maxlat_per_CPU[m].iter
                  <<" task "<<maxlat_per_CPU[m].task<<endl ;
#endif
#if 0           
           // update global aggregated statistics, i.e. maxdiff_global, 
           // mindiff_global, maxratio_global and minratio_global
           if (maxdiff_per_CPU[m].diff > maxdiff_global.diff) 
               maxdiff_global = maxdiff_per_CPU[m] ;
               
           if (mindiff_global.diff < -MIN_DOUBLE_CONST || 
                    mindiff_per_CPU[m].diff < mindiff_global.diff)
               mindiff_global = mindiff_per_CPU[m] ;

           if (maxratio_per_CPU[m].ratio > maxratio_global.ratio)
               maxratio_global = maxratio_per_CPU[m] ;

           if (maxratio_UJ_per_CPU[m].ratio_UJ > maxratio_UJ_global.ratio_UJ)
               maxratio_UJ_global = maxratio_UJ_per_CPU[m] ;

           if (minratio_global.ratio < 0 || 
               minratio_per_CPU[m].ratio < minratio_global.ratio)
               minratio_global = minratio_per_CPU[m] ;
#endif

           if (maxlat_per_CPU[m].maxlat > maxlat_global.maxlat)
               maxlat_global = maxlat_per_CPU[m] ;

	   iter-- ; // here current iteration counter is one unit too high
	   save_state() ; // save state after last iteration for this system
	   iter = 1 ; // reset current iteration counter for next system
       } // end of main loop

       // log global aggregated statistics, i.e.
       // maxdiff_global, mindiff_global, maxratio_global and minratio_global
       DEB(cout<<"\nAggregated global statistics"<<endl);
#if 0
       DEB(cout<<"\tMinimum global diff "<<mindiff_global.diff
               <<" occurred with "<<mindiff_global.CPU<<" CPUs, at iteration "
               <<mindiff_global.iter<<","<<endl<<"\tfor task "
               <<mindiff_global.task<<endl ) ;
       DEB(cout<<"\tMaximum global diff "<<maxdiff_global.diff
               <<" occurred with "<<maxdiff_global.CPU<<" CPUs, at iteration "
               <<maxdiff_global.iter<<","<<endl<<"\tfor task "
               <<maxdiff_global.task<<endl ) ;
       DEB(cout<<"\tMinimum global ratio "<<minratio_global.ratio
               <<" occurred with "<<minratio_global.CPU<<" CPUs, at iteration "
               <<minratio_global.iter<<","<<endl<<"\tfor task "
               <<minratio_global.task<<endl ) ;
       DEB(cout<<"\tMaximum global ratio "<<maxratio_global.ratio
               <<" occurred with "<<maxratio_global.CPU<<" CPUs, at iteration "
               <<maxratio_global.iter<<","<<endl<<"\tfor task "
               <<maxratio_global.task<<endl ) ;
       DEB(cout<<"\tMaximum global ratio UJ "<<maxratio_UJ_global.ratio_UJ
               <<" occurred with "<<maxratio_UJ_global.CPU
	       <<" CPUs, at iteration "
               <<maxratio_UJ_global.iter<<","<<endl<<"\tfor task "
               <<maxratio_UJ_global.task<<endl ) ;
#endif
       DEB(cout<<"\tMaximum global lateness "<<maxlat_global.maxlat
               <<" occurred with "<<maxlat_global.CPU<<" CPUs, at iteration "
               <<maxlat_global.iter<<","<<endl<<"\tfor task "
               <<maxlat_global.task<<endl ) ;
#if 0
       Out.max<<"\nMin_diff_global "<<mindiff_global.diff<<" CPUs "
              <<mindiff_global.CPU<<" iteration "<<mindiff_global.iter
              <<" task "<<mindiff_global.task<<endl ;
       Out.max<<"\nMax_diff_global "<<maxdiff_global.diff<<" CPUs "
              <<maxdiff_global.CPU<<" iteration "<<maxdiff_global.iter
              <<" task "<<maxdiff_global.task<<endl ;
       Out.max<<"\nMin_ratio_global "<<minratio_global.diff<<" CPUs "
              <<minratio_global.CPU<<" iteration "<<minratio_global.iter
              <<" task "<<minratio_global.task<<endl ;
       Out.max<<"\nMax_ratio_global "<<maxratio_global.diff<<" CPUs "
              <<maxratio_global.CPU<<" iteration "<<maxratio_global.iter
              <<" task "<<maxratio_global.task<<endl ;
       Out.max<<"\nMax_lateness_global "<<maxlat_global.maxlat<<" CPUs "
              <<maxlat_global.CPU<<" iteration "<<maxlat_global.iter
              <<" task "<<maxlat_global.task<<endl ;
#endif
       // per system summaries
       M-- ; // to get a coherent summary
       per_system_summaries(
#if 0
Out
#endif
) ;
       
#if 0
       // close files       
       Out.mean.close();
       Out.max.close();
#endif
    } catch (BaseExc &e) {
       cout << e.what() << endl;
    } catch (std::exception &e) {
	   cout << "Exception: " << e.what() << " at time: " << 
            Simulation::getTime() << endl;
    } catch (const char *s) {
       cout << s << endl;
    } catch (...) {
	   cout << "Unknown error" << endl;
    }
  /* Writing a seed for the next random numbers generator */
//   ofstream o_rand_num;
//   o_rand_num.open("rand_gen.txt");
//   o_rand_num << rand();
//   o_rand_num.close();
} // end of main

